(function (GCN) {

	'use strict';

	/**
	 * Only grows, never shrinks.
	 * @private
	 * @type {number}
	 */
	var uniqueIdCounter = 0;

	/**
	 * Generates a unique id with an optional prefix.
	 *
	 * The returned value is only unique among other returned values,
	 * not globally.
	 *
	 * @public
	 * @param {string}
	 *        Optional prefix for the id to be generated.
	 * @return {string}
	 *        Never the same string more than once.
	 */
	function uniqueId(prefix) {
		return (prefix || '') + (++uniqueIdCounter);
	}

	function escapePropertyName(name) {
		return name.replace(/\./g, '\\.');
	}

	/**
	 * A regular expression to identify executable script tags.
	 *
	 * @type {RegExp}
	 */
	var rgxpScriptType = /\/(java|ecma)script/i;

	/**
	 * A regular expression to find script tags.
	 *
	 * @type {RegExp}
	 */
	var rgxpScriptTag = new RegExp('<script(\\s[^>]*?)?>', 'ig');

	/**
	 * A regular expression to find script types.
	 *
	 * @type {RegExp}
	 */
	var rgxpType = new RegExp(
		' type\\s*=\\s*'
			+ '[\\"\\\']'
			+ '([^\\"\\\']*[^\\s][^\\"\\\']*)'
			+ '[\\"\\\']',
		'i'
	);

	var rand = Math.random().toString().replace('.', '');

	/**
	 * Places sentinal strings in front of every executable script tag.
	 *
	 * @param {string} html HTML markup
	 * @return {string} HTML string with marked <script> tags.
	 */
	function markScriptTagLocations(html) {
		var i = 0;
		return html.replace(rgxpScriptTag, function (str, substr, offset) {
			var type = substr && substr.match(rgxpType);
			if (!type || rgxpScriptType.test(type)) {
				return rand + (i++) + str;
			}
			return str;
		});
	}

	/**
	 * Masks a <script> tag's `type' attribute by replacing it with a random
	 * string.
	 *
	 * Masking script tags is don to prevent them from being handled specially
	 * by jQuery which removes javascript/ecmascript tags when appending DOM
	 * elements into the document, but executes them.
	 *
	 * unmakeScriptType() does the reverse of this.
	 *
	 * @param {jQuery.<HTMLElement>} $script jQuery unit set containing the
	 *                                       <script> tag that is to have its
	 *                                       type attribute masked.
	 */
	function maskScriptType($script) {
		var type = $script.attr('type');
		$script.attr('type', rand).attr('data-origtype', type);
	}

	/**
	 * Restores a <script> tag's original type attribute value, if it had been
	 * masked using maskScriptType().
	 *
	 * Essentially the reverse of maskScriptType().
	 *
	 * @param {jQuery.<HTMLElement>} $script jQuery unit set containing the
	 *                                       <script> tag that is to have its
	 *                                       type attribute unmasked.
	 */
	function unmaskScriptType($script) {
		var orig = $script.attr('data-origtype');
		if (typeof orig === 'string') {
			$script.attr('type', orig).removeAttr('data-origtype');
		} else {
			$script.removeAttr('type');
		}
	}

	/**
	 * Replaces the type attribute of <script> tags with a value that will
	 * protect them from being specially handled by jQuery.
	 *
	 * @param {string} html Markup
	 * @return {string} Markup with <script> tags protected from jQuery.
	 */
	function protectScriptTags(html, $scripts) {
		var i;
		var $script;
		var type;
		for (i = 0; i < $scripts.length; i++) {
			$script = $scripts.eq(i);
			type = $script.attr('type');
			if (!type || rgxpScriptType.test(type)) {
				maskScriptType($script);
				html = html.replace(rand + i, $script[0].outerHTML);
				unmaskScriptType($script);
			}
		}
		return html;
	}

	/**
	 * Restores the type attribute for <script> tags that have been processed
	 * via protectScriptTags().
	 *
	 * @param {jQuery.<HTMLElement>} $element Root HTMLElement in which to
	 *                                        restore <script> tags.
	 */
	function restoreScriptTagTypes($element) {
		var $scripts = $element.find('script[type="' + rand + '"]');
		var $script;
		var i;
		for (i = 0; i < $scripts.length; i++) {
			$script = $scripts.eq(i);
			$script.removeClass(rand + i);
			unmaskScriptType($script);
		}
	}

	/**
	 * Joins the innerHTML of multiple elements.
	 *
	 * @param {jQuery.<HTMLElement>} $elements
	 */
	function joinContents($elements) {
		var contents = '';
		$elements.each(function () {
			contents += jQuery(this).html();
		});
		return contents;
	}

	/**
	 * Inserts the inner HTML of the given HTML markup while preserving
	 * <script> tags, and still allowing jQuery to execute them.
	 *
	 * @param {jQuery.<HTMLElement>} $element
	 * @param {string} html
	 */
	function insertInnerHTMLWithScriptTags($element, html) {
		if (!rgxpScriptTag.test(html)) {
			$element.html(jQuery(html).contents());
			return;
		}
		var marked = markScriptTagLocations(html);
		var $scripts = jQuery(html).filter('script');
		var $html = jQuery(marked);
		var contents = joinContents($html.filter(':not(script)'));
		contents = protectScriptTags(contents, $scripts);
		$element.html(contents);

		// Trap script errors originating from rendered tags and log it on the
		// console.
		try {
			$element.append($scripts);
		} catch (ex) {
			var _console = 'console'; // Avoid jslint warnings
			if (typeof window[_console] === 'function') {
				window[_console].error(ex);
			}
		}

		restoreScriptTagTypes($element);
	}

	/**
	 * Merge class names from one element into another.
	 *
	 * The merge result will be a unqiue set of space-seperated class names.
	 *
	 * @param {jQuery.<HTMLElement>} $first jQuery unit set containing the DOM
	 *                                      whose class names are to be merged.
	 * @param {jQuery.<HTMLElement>} $second jQuery unit set containing the DOM
	 *                                       whose class names are to be merged.
	 * @return {string} The merge result of the merge: a unqiue set of
	 *                  space-seperated class names.
	 */
	function mergeClassNames($first, $second) {
		var first = ($first.attr('class') || '').split(' ');
		var second = ($second.attr('class') || '').split(' ');
		var names = first.concat(second).sort();
		var i;
		for (i = 1; i < names.length; i++) {
			if (names[i] === names[i - 1]) {
				names.splice(i--, 1);
			}
		}
		return names.join(' ');
	}

	/**
	 * Creates a map of attributes merged from their value in $from with their
	 * value in $to.
	 *
	 * Unlike other attributes, which are simply copied from $from into $to,
	 * class names are treaded specially to produce unique set of
	 * space-seperated class names.
	 *
	 * @param {jQuery.<HTMLElement>} $to jQuery unit set containing the DOM
	 *                                   element which should receive the
	 *                                   merged attributes.
	 * @param {jQuery.<HTMLElement>} $from jQuery unit set containing the DOM
	 *                                     element whose attributes will be
	 *                                     merged into the other.
	 * @param {object<string, string>} A associate array of attributes.
	 */
	function mergeAttributes($to, $from) {
		var to = $to[0].attributes;
		var from = $from[0].attributes;
		var i;
		var attr = {};
		for (i = 0; i < from.length; i++) {
			attr[from[i].name] = ('class' === from[i].name)
			                   ? mergeClassNames($to, $from)
			                   : $from.attr(from[i].name);
		}
		return attr;
	}

	/**
	 * Renders the given HTML string onto (not into) an DOM element.
	 * Does nearly the equivelent of changes the elements' outerHTML.
	 *
	 * @param {jQuery.<HTMLElement>} $element jQuery unit set containing the
	 *                                        DOM element we wish to render the
	 *                                        given html content onto.
	 * @param {string} html HTML content which will become give to the element.
	 */
	function renderOnto($element, html) {
		insertInnerHTMLWithScriptTags($element, html);
		var attr = mergeAttributes($element, jQuery(html));
		var name;
		for (name in attr) {
			if (attr.hasOwnProperty(name)) {
				$element.attr(name, attr[name]);
			}
		}
	}

	GCN.uniqueId = uniqueId;
	GCN.escapePropertyName = escapePropertyName;
	GCN.renderOnto = renderOnto;

}(GCN));
